"""
parser and parsing-related tools

Design
------

Template string is split into tokens and the tokens are combined into
nodes. Parse tree is a nodelist. TextNode and ExpressionNode are simple
nodes and for-loop, if-loop, etc. are block nodes, which contain multi-
ple child nodes.

Each node can emit some python string. python string emitted by the root
node is validated for safeeval and executed using python in the given
environment.

Enough care is taken to make sure the generated code and the template
has line to line match, so that the error messages can point to exact
line number in template. (It still doesn't work in some cases.)

Grammar
-------

    template -> defwith sections
    defwith -> "$def with (" arguments ")" | ""
    sections -> section*
    section -> block | assignment | line

    assignment -> "$ " <assignment expression>
    line -> (text|expr)*
    text -> <any characters other than $>
    expr -> "$" pyexpr | "$(" pyexpr ")" | "${" pyexpr "}"
    pyexpr -> <python expression>

"""

import collections
import re
import token
import tokenize

__all__ = ["ParseError", "SecurityError", "ForLoop"]


class Error(Exception):

    """"""


class ParseError(Error):

    """"""


class SecurityError(Error):

    """the template seems to be trying to do something naughty"""


_indent = " " * 4


def splitline(text):
    r"""
    splits `text` at first newline (modified string `partition`)

        >>> splitline("foo\nbar")
        ('foo\n', 'bar')
        >>> splitline('foo')
        ('foo', '')
        >>> splitline("")
        ('', '')

    """
    before, partition, after = text.partition("\n")
    return before + partition, after


class Node:

    """ """


class DefWith(Node):

    """ """

    def __init__(self, defwith, suite):
        if defwith:
            self.defwith = defwith.replace("with", "template__") + ":"
            offset = 4  # encoding, lineoffset__, loop & self
        else:
            self.defwith = "def template__():"
            offset = 5  # encoding, template__, lineoffset__, loop & self
        self.defwith += "\n    lineoffset__ = -{0}".format(offset)
        self.defwith += "\n    loop = ForLoop()"
        self.defwith += "\n    self = TemplateResult()"
        self.defwith += "\n    e_ = self.extend"
        self.suite = suite

    def emit(self, indent):
        return "\n".join(
            (self.defwith, self.suite.emit(indent + _indent), "    return self")
        )

    def __repr__(self):
        return "<defwith: {0}, {1}>".format(self.defwith, self.suite)


class Var(Node):

    """ """

    def __init__(self, name, value):
        self.name = name
        self.value = value

    def emit(self, indent, text_indent):
        return indent + "self[{0}] = {1}\n".format(repr(self.name), self.value)

    def __repr__(self):
        return "<var: {0} = {1}>".format(self.name, self.value)


class Text(Node):

    """ """

    def __init__(self, value):
        self.value = value

    def emit(self, indent, begin_indent=""):
        return repr(self.value)

    def __repr__(self):
        return "t" + repr(self.value)


class Expression:

    """ """

    def __init__(self, value, escape=True):
        self.value = value.strip()
        if value.startswith("{") and value.endswith("}"):  # ${...} to $(...)
            self.value = "(" + self.value[1:-1] + ")"
        self.escape = escape

    def emit(self, indent, begin_indent=""):
        return "escape_({0}, {1})".format(self.value, bool(self.escape))

    def __repr__(self):
        if self.escape:
            escape = ""
        else:
            escape = ":"
        return "${0}{1}".format(escape, self.value)


class Assignment:

    """ """

    def __init__(self, code):
        self.code = code

    def emit(self, indent, begin_indent=""):
        return indent + self.code + "\n"

    def __repr__(self):
        return "<assignment: {0}>".format(repr(self.code))


class Line:

    """ """

    def __init__(self, nodes):
        self.nodes = nodes

    def emit(self, indent, text_indent="", name=""):
        text = [node.emit("") for node in self.nodes]
        if text_indent:
            text = [repr(text_indent)] + text
        return indent + "e_([{0}])\n".format(", ".join(text))

    def __repr__(self):
        return "<line: {0}>".format(repr(self.nodes))


class Code:

    """ """

    def __init__(self, stmt, block, begin_indent=""):
        # compensate one line for $code:
        self.code = "\n" + block

    def emit(self, indent, text_indent=""):
        rx = re.compile("^", re.M)
        return rx.sub(indent, self.code).rstrip(" ")

    def __repr__(self):
        return "<code: {0}>".format(repr(self.code))


class Statement:

    """ """

    def __init__(self, stmt):
        self.stmt = stmt

    def emit(self, indent, begin_indent=""):
        return indent + self.stmt

    def __repr__(self):
        return "<stmt: {0}>".format(repr(self.stmt))


class Suite:

    """
    suite is a list of sections

    """

    def __init__(self, sections):
        self.sections = sections

    def emit(self, indent, text_indent=""):
        content = []
        for section in self.sections:
            content.append(section.emit(indent, text_indent))
        return "\n" + "".join(content)

    def __repr__(self):
        return repr(self.sections)


class Block:

    """ """

    def __init__(self, stmt, block, begin_indent=""):
        self.stmt = stmt
        self.suite = Parser().read_suite(block)
        self.begin_indent = begin_indent

    def emit(self, indent, text_indent=""):
        text_indent = self.begin_indent + text_indent
        out = self.suite.emit(indent + _indent, text_indent)
        return indent + self.stmt + out

    def __repr__(self):
        return "<block: {0}, {1}>".format(repr(self.stmt), repr(self.suite))


class Def(Block):

    """ """

    def __init__(self, *a, **kw):
        Block.__init__(self, *a, **kw)
        code = Code("", "")
        code.code = "self = TemplateResult(); e_ = self.extend\n"
        self.suite.sections.insert(0, code)
        code = Code("", "")
        code.code = "return self\n"
        self.suite.sections.append(code)

    def emit(self, indent, text_indent=""):
        text_indent = self.begin_indent + text_indent
        out = self.suite.emit(indent + _indent, text_indent)
        return indent + "lineoffset__ -= 3\n" + indent + self.stmt + out


class If(Block):

    pass


class Else(Block):

    pass


class Elif(Block):

    pass


class For(Block):

    """ """

    def __init__(self, stmt, block, begin_indent=""):
        self.original_stmt = stmt
        tok = Tokenizer(stmt)
        tok.consume_till("in")
        a = stmt[: tok.index]  # for i in
        b = stmt[tok.index : -1]  # rest of for stmt excluding :
        stmt = a + " loop.setup(" + b.strip() + "):"
        Block.__init__(self, stmt, block, begin_indent)

    def __repr__(self):
        return "<block: {0}, {1}>".format(repr(self.original_stmt), repr(self.suite))


class ForLoop:

    """
    wrapper for expression in for stament to support loop.xxx helpers

            >>> loop = ForLoop()
            >>> for x in loop.setup(["a", "b", "c"]):
            ...     loop.index, loop.revindex, loop.parity, x
            ...
            (1, 3, 'odd', 'a')
            (2, 2, 'even', 'b')
            (3, 1, 'odd', 'c')
            >>> loop.index
            Traceback (most recent call last):
                    ...
            AttributeError: index

    """

    def __init__(self):
        self._ctx = None

    def __getattr__(self, name):
        if self._ctx is None:
            raise AttributeError(name)
        else:
            return getattr(self._ctx, name)

    def setup(self, seq):
        self._push()
        return self._ctx.setup(seq)

    def _push(self):
        self._ctx = ForLoopContext(self, self._ctx)

    def _pop(self):
        self._ctx = self._ctx.parent


class ForLoopContext:

    """
    stackable context for ForLoop to support nested for loops

    """

    index0 = property(lambda self: self.index - 1)
    first = property(lambda self: self.index == 1)
    last = property(lambda self: self.index == self.length)
    odd = property(lambda self: self.index % 2 == 1)
    even = property(lambda self: self.index % 2 == 0)
    parity = property(lambda self: ["odd", "even"][self.even])
    revindex0 = property(lambda self: self.length - self.index)
    revindex = property(lambda self: self.length - self.index + 1)

    def __init__(self, forloop, parent):
        self._forloop = forloop
        self.parent = parent

    def setup(self, seq):
        try:
            self.length = len(seq)
        except:
            self.length = 0
        self.index = 0
        for a in seq:
            self.index += 1
            yield a
        self._forloop._pop()


class Parser:

    """ """

    keywords = ("pass", "break", "continue", "return")
    statement_nodes = {
        "for": For,
        "while": Block,
        "if": If,
        "elif": Elif,
        "else": Else,
        "def": Def,
        "code": Code,
    }

    def __init__(self):
        self.statement_nodes = self.statement_nodes
        self.keywords = self.keywords

    def parse(self, text, name="<template>"):
        self.text = text
        self.name = name
        defwith, text = self.read_defwith(text)
        suite = self.read_suite(text)
        return DefWith(defwith, suite)

    def read_defwith(self, text):
        if text.startswith("$def with"):
            defwith, text = splitline(text)
            defwith = defwith[1:].strip()  # strip $ and spaces
            return defwith, text
        else:
            return "", text

    def read_section(self, text):
        r"""
        reads one section from the given text

            section -> block | assignment | line

            >>> read_section = Parser().read_section
            >>> read_section("foo\nbar\n")
            (<line: [t'foo\n']>, 'bar\n')
            >>> read_section("$ a = b + 1\nfoo\n")
            (<assignment: 'a = b + 1'>, 'foo\n')

            # TODO >>> read_section("$for i in range(10):\n  A $i\nB")
            # TODO (<block: 'for i in range(10):', [<line: [t'A ', $i, t'\n']>]>, 'B')

        """
        if text.lstrip(" ").startswith("$"):
            index = text.index("$")
            begin_indent, text2 = text[:index], text[index + 1 :]
            ahead = self.python_lookahead(text2)
            if ahead == "var":
                return self.read_var(text2)
            elif ahead in self.statement_nodes:
                return self.read_block_section(text2, begin_indent)
            elif ahead in self.keywords:
                return self.read_keyword(text2)
            elif ahead.strip() == "":
                # assignments starts with a space after $
                # ex: $ a = b + 2
                return self.read_assignment(text2)
        return self.readline(text)

    def read_var(self, text):
        r"""
        reads a `var` statement

        >>> read_var = Parser().read_var
        >>> read_var("var x=10\nfoo")
        (<var: x = 10>, 'foo')
        >>> read_var("var x: hello $name\nfoo")
        (<var: x = join_('hello ', escape_(name, True))>, 'foo')

        """
        line, text = splitline(text)
        tokens = self.python_tokens(line)
        if len(tokens) < 4:
            raise SyntaxError("Invalid var statement")
        name = tokens[1]
        sep = tokens[2]
        value = line.split(sep, 1)[1].strip()
        if sep == "=":
            pass  # no need to process value
        elif sep == ":":
            # FIXME hack for backward-compatability
            if tokens[3] == "\n":  # multi-line var statement
                block, text = self.read_indented_block(text, "        ")
                lines = [self.readline(x)[0] for x in block.splitlines()]
                _nodes = []
                for x in lines:
                    _nodes.extend(x.nodes)
                    _nodes.append(Text("\n"))
            else:  # single-line var statement
                linenode, _ = self.readline(value)
                _nodes = linenode.nodes
            parts = [node.emit("") for node in _nodes]
            value = "join_({0})".format(", ".join(parts))
        else:
            raise SyntaxError("Invalid var statement")
        return Var(name, value), text

    def read_suite(self, text):
        r"""
        reads section by section till end of text

        >>> read_suite = Parser().read_suite
        >>> read_suite("hello $name\nfoo\n")
        [<line: [t'hello ', $name, t'\n']>, <line: [t'foo\n']>]

        """
        sections = []
        while text:
            section, text = self.read_section(text)
            sections.append(section)
        return Suite(sections)

    def readline(self, text):
        r"""
        reads one line from the text

        Newline is supressed if the line ends with \.

        >>> readline = Parser().readline
        >>> readline("hello $name!\nbye!")
        (<line: [t'hello ', $name, t'!\n']>, 'bye!')
        >>> readline("hello $name!\\\nbye!")
        (<line: [t'hello ', $name, t'!']>, 'bye!')
        >>> readline("$f()\n\n")
        (<line: [$f(), t'\n']>, '\n')

        """
        line, text = splitline(text)
        # supress new line if line ends with \
        if line.endswith("\\\n"):
            line = line[:-2]
        _nodes = []
        while line:
            node, line = self.read_node(line)
            _nodes.append(node)
        return Line(_nodes), text

    def read_node(self, text):
        r"""
        reads a node from the given text and returns the node and
        remaining text

            >>> read_node = Parser().read_node
            >>> read_node("hello $name")
            (t'hello ', '$name')
            >>> read_node("$name")
            ($name, '')

        """
        if text.startswith("$$"):
            return Text("$"), text[2:]
        elif text.startswith("$#"):  # comment
            line, text = splitline(text)
            return Text("\n"), text
        elif text.startswith("$"):
            text = text[1:]  # strip $
            if text.startswith(":"):
                escape = False
                text = text[1:]  # strip `:`
            else:
                escape = True
            return self.read_expr(text, escape=escape)
        else:
            return self.read_text(text)

    def read_text(self, text):
        r"""
        reads a text node from the given text

            >>> read_text = Parser().read_text
            >>> read_text("hello $name")
            (t'hello ', '$name')

        """
        index = text.find("$")
        if index < 0:
            return Text(text), ""
        else:
            return Text(text[:index]), text[index:]

    def read_keyword(self, text):
        line, text = splitline(text)
        return Statement(line.strip() + "\n"), text

    def read_expr(self, text, escape=True):
        """
        reads a python expression from the text and returns the expression
        and remaining text

            expr -> simple_expr | paren_expr
            simple_expr -> id extended_expr
            extended_expr -> attr_access | paren_expr extended_expr | ""
            attr_access -> dot id extended_expr
            paren_expr -> [ tokens ] | ( tokens ) | { tokens }

            >>> read_expr = Parser().read_expr
            >>> read_expr("name")
            ($name, '')
            >>> read_expr("a.b and c")
            ($a.b, ' and c')
            >>> read_expr("a. b")
            ($a, '. b')
            >>> read_expr("name</h1>")
            ($name, '</h1>')
            >>> read_expr("(limit)ing")
            ($(limit), 'ing')
            >>> read_expr('a[1, 2][:3].f(1+2, "weird string[).", 3 + 4) done')
            ($a[1, 2][:3].f(1+2, "weird string[).", 3 + 4), ' done')

        """
        parens = {"(": ")", "[": "]", "{": "}"}

        def simple_expr():
            identifier()
            extended_expr()

        def identifier():
            next(tokens)

        def extended_expr():
            lookahead = tokens.lookahead()
            if lookahead is None:
                return
            elif lookahead.value == ".":
                attr_access()
            elif lookahead.value in parens:
                paren_expr()
                extended_expr()
            else:
                return

        def attr_access():
            tokens.lookahead()
            # XXX dot = tokens.lookahead()
            if tokens.lookahead2().type == token.NAME:
                next(tokens)  # XXX can remove comment? -- # consume dot
                identifier()
                extended_expr()

        def paren_expr():
            begin = next(tokens).value
            end = parens[begin]
            while True:
                if tokens.lookahead().value in parens:
                    paren_expr()
                else:
                    t = next(tokens)
                    if t.value == end:
                        break
            return

        def get_tokens(text):
            """
            tokenizes `text` using python tokenizer

            Python tokenizer ignores spaces, but they might be important in
            some cases. This function introduces dummy space tokens when it
            identifies any ignored space. Each token is a `NamedTuple` with
            attributes type, value, begin and end.

            """
            readline = iter([text]).__next__
            end = None
            for t in tokenize.generate_tokens(readline):
                t = Token(type=t[0], value=t[1], begin=t[2], end=t[3])
                if end is not None and end != t.begin:
                    _, x1 = end
                    _, x2 = t.begin
                    yield Token(type=-1, value=text[x1:x2], begin=end, end=t.begin)
                end = t.end
                yield t

        class BetterIter:

            """
            iterator-like object with support for 2 look-aheads

            """

            def __init__(self, items):
                self.iteritems = iter(items)
                self.items = []
                self.position = 0
                self.current_item = None

            def lookahead(self):
                if len(self.items) <= self.position:
                    self.items.append(self._next())
                return self.items[self.position]

            def _next(self):
                try:
                    return next(self.iteritems)
                except StopIteration:
                    return None

            def lookahead2(self):
                if len(self.items) <= self.position + 1:
                    self.items.append(self._next())
                return self.items[self.position + 1]

            def __next__(self):
                self.current_item = self.lookahead()
                self.position += 1
                return self.current_item

        tokens = BetterIter(get_tokens(text))
        if tokens.lookahead().value in parens:
            paren_expr()
        else:
            simple_expr()
        row, col = tokens.current_item.end
        return Expression(text[:col], escape=escape), text[col:]

    def read_assignment(self, text):
        r"""
        reads assignment statement from text

        >>> read_assignment = Parser().read_assignment
        >>> read_assignment("a = b + 1\nfoo")
        (<assignment: 'a = b + 1'>, 'foo')

        """
        line, text = splitline(text)
        return Assignment(line.strip()), text

    def python_lookahead(self, text):
        """
        returns the first python token from the given text

        >>> python_lookahead = Parser().python_lookahead
        >>> python_lookahead("for i in range(10):")
        'for'
        >>> python_lookahead("else:")
        'else'
        >>> python_lookahead(" x = 1")
        ' '

        """
        readline = iter([text]).__next__
        tokens = tokenize.generate_tokens(readline)
        return next(tokens)[1]

    def python_tokens(self, text):
        readline = iter([text]).__next__
        tokens = tokenize.generate_tokens(readline)
        return [t[1] for t in tokens]

    def read_indented_block(self, text, indent):
        r"""
        reads a block of text

        A block is what typically follows a `for` or `if` statement. It can
        be in the same line as that of the statement or an indented block.

        >>> read_indented_block = Parser().read_indented_block
        >>> read_indented_block("    a\n    b\nc", "    ")
        ('a\nb\n', 'c')
        >>> read_indented_block("    a\n        b\n    c\nd", "    ")
        ('a\n    b\nc\n', 'd')
        >>> read_indented_block("    a\n\n        b\nc", "    ")
        ('a\n\n    b\n', 'c')

        """
        if indent == "":
            return "", text
        block = ""
        while text:
            line, text2 = splitline(text)
            if line.strip() == "":
                block += "\n"
            elif line.startswith(indent):
                block += line[len(indent) :]
            else:
                break
            text = text2
        return block, text

    def read_statement(self, text):
        r"""
        reads a python statement

            >>> read_statement = Parser().read_statement
            >>> read_statement("for i in range(10): hello $name")
            ('for i in range(10):', ' hello $name')

        """
        tok = Tokenizer(text)
        tok.consume_till(":")
        return text[: tok.index], text[tok.index :]

    def read_block_section(self, text, begin_indent=""):
        r"""
        TODO

            >>> read_block_section = Parser().read_block_section
            >>> read_block_section("for i in range(10): hi $i\nB")
            (<block: 'for i in range(10):', [<line: [t'hi ', $i, t'\n']>]>, 'B')
            >>> read_block_section("for i in range(10):\n      hi $i\n    B",
            ...                    begin_indent="    ")
            (<block: 'for i in range(10):', [<line: [t'hi ', $i, t'\n']>]>, '    B')
            >>> read_block_section("for i in range(10):\n    hi $i\nB")
            (<block: 'for i in range(10):', [<line: [t'hi ', $i, t'\n']>]>, 'B')

        """
        line, text = splitline(text)
        stmt, line = self.read_statement(line)
        keyword = self.python_lookahead(stmt)
        # if there is some thing left in the line
        if line.strip():
            block = line.lstrip()
        else:

            def find_indent(text):
                rx = re.compile("    +")
                match = rx.match(text)
                first_indent = match and match.group(0)
                return first_indent or ""

            # find the indentation of the block by looking at the first line
            first_indent = find_indent(text)[len(begin_indent) :]
            # TODO fix this special case
            if keyword == "code":
                indent = begin_indent + first_indent
            else:
                indent = begin_indent + min(first_indent, _indent)
            block, text = self.read_indented_block(text, indent)
        return self.create_block_node(keyword, stmt, block, begin_indent), text

    def create_block_node(self, keyword, stmt, block, begin_indent):
        if keyword in self.statement_nodes:
            return self.statement_nodes[keyword](stmt, block, begin_indent)
        else:
            raise ParseError("Unknown statement: {0}".format(repr(keyword)))


Token = collections.namedtuple("Token", "type, value, begin, end")


class Tokenizer:

    """
    utility wrapper over `tokenize` token generator

    """

    def __init__(self, text):
        self.text = text
        readline = iter([text]).__next__
        self.tokens = tokenize.generate_tokens(readline)
        self.index = 0

    def consume_till(self, delim):
        """
        consume tokens until colon

            >>> tok = Tokenizer("for i in range(10): hello $i")
            >>> tok.consume_till(":")
            >>> tok.text[:tok.index]
            'for i in range(10):'
            >>> tok.text[tok.index:]
            ' hello $i'

        """
        try:
            while True:
                t = next(self)
                if t.value == delim:
                    break
                elif t.value == "(":
                    self.consume_till(")")
                elif t.value == "[":
                    self.consume_till("]")
                elif t.value == "{":
                    self.consume_till("}")
                # if end of line is found, it is an exception.
                # Since there is no easy way to report the line number,
                # leave the error reporting to the python parser later
                # TODO this should be fixed
                if t.value == "\n":
                    break
        except:
            # error_msg = "Expected {}, found end of line."
            # raise ParseError(error_msg.format(repr(delim)))
            # Raising ParseError doesn"t show the line number. If this error
            # is ignored then it will be caught when compiling the python code
            return

    def __next__(self):
        type, t, begin, end, line = next(self.tokens)
        row, col = end
        self.index = col
        return Token(type=type, value=t, begin=begin, end=end)



"""
Supported machine hosts and domain registrars.

Supported hosts: DigitalOcean, Linode, Hetzner
Supported registrars: Dynadot, Name.com

"""

import logging
import time

import lxml.etree
import requests


class TokenError(Exception):
    """Bad auth token."""


class DomainExistsError(Exception):
    """Domain already exists."""


class Provider:
    """A service provider."""

    def __init__(self, token=None):
        """Store the access token."""
        self.token = token


class Host(Provider):
    """A machine host."""


class Registrar(Provider):
    """A domain registrar."""


class DigitalOcean(Host):
    """
    [DigitalOcean][0] client.

    [0]: https://digitalocean.com

    """

    endpoint = "https://api.digitalocean.com/v2"

    def __init__(self, token):
        """Set up authenticated session."""
        self.session = requests.Session()
        self.session.headers["Authorization"] = f"Bearer {token}"
        super().__init__(token)

    def get_keys(self):
        """Return all SSH keys."""
        return self._request("get", "account/keys")

    def add_key(self, name, key_data):
        """Add an SSH key."""
        return self._request(
            "post", "account/keys", json={"name": name, "public_key": key_data}
        )["ssh_key"]

    @property
    def machines(self):
        """Return all machines."""
        machines = []
        for machine in self._request("get", "droplets")["droplets"]:
            machine.pop("region")
            machines.append(machine)
        return machines

    def create_machine(
        self,
        name,
        region="sfo3",
        size="s-1vcpu-512mb-10gb",  # $4/mo
        # size="s-1vcpu-1gb-intel",  # $7/mo
        image="debian-11-x64",
        ssh_keys=None,
    ):
        """Create a machine."""
        # machine = None
        machine_id = self._request(
            "post",
            "droplets",
            json={
                "name": name,
                "region": region,
                "size": size,
                "image": image,
                "ssh_keys": ssh_keys,
            },
        )["droplet"]["id"]
        while (
            self._request("get", f"droplets/{machine_id}/actions")["actions"][0][
                "status"
            ]
            == "in-progress"
        ):
            time.sleep(1)
        tries = 5
        while tries:
            machine = self.get_machine(machine_id)["droplet"]
            if machine["networks"]["v4"]:
                break
            time.sleep(2)
            tries -= 1
        return machine

    def get_machine(self, machine_id):
        """Get a machine."""
        return self._request("get", f"droplets/{machine_id}")

    def delete_machine(self, machine_id):
        """Delete a machine."""
        return self._request("delete", f"droplets/{machine_id}")

    def _request(self, method, path, **kwargs):
        response = getattr(self.session, method)(f"{self.endpoint}/{path}", **kwargs)
        if response.status_code == 422:
            logging.error(f"DigitalOcean:{response.json()['message']}")
        return response.json()


class Linode(Host):
    """
    [Linode][0] client.

    [0]: https://linode.com

    """


class Hetzner(Host):
    """
    [Hetzner][0] client.

    [0]: https://hetzner.com

    """


class Dynadot(Registrar):
    """
    [Dynadot][0] client.

    [0]: https://dynadot.com

    """

    endpoint = "https://api.dynadot.com/api3.xml"

    @property
    def domains(self):
        """List currently registered domains."""
        response = self._request("list_domain")
        domains = []
        for domain in response.cssselect("Domain"):
            name = domain.cssselect("Name")[0].text
            expiration = domain.cssselect("Expiration")[0].text
            domains.append((name, expiration))
        return sorted(domains)

    def create_record(self, domain, record, subdomain=""):
        """Set DNS record for given domain."""
        # TODO set low ttl
        command = "set_dns2"
        record_type = "a"
        if subdomain:
            return self._request(
                command,
                domain=domain,
                main_record_type0=record_type,
                main_record0=record,
                subdomain0=subdomain,
                sub_record_type0=record_type,
                sub_record0=record,
            )
        return self._request(
            command, domain=domain, main_record_type0=record_type, main_record0=record
        )

    def search(self, *domains):
        """Search for available of domains."""
        domain_params = {
            "domain{}".format(n): domain for n, domain in enumerate(domains)
        }
        response = self._request(show_price="1", **domain_params)
        results = {}
        for result in response:
            # if len(result[0]) == 5:
            # data = {"price": result[0][4].text}
            # results[result[0][1].text] = data
            available = False if result[0].find("Available").text == "no" else True
            price = result[0].find("Price")
            if price is None:
                price = 0
            else:
                if " in USD" in price.text:
                    price = float(price.text.partition(" ")[0])
                else:
                    price = "?"
            results[result[0].find("DomainName").text] = (available, price)
        return results

    def register(self, domain, duration=1):
        """Register domain."""
        return self._request("register", domain=domain, duration=duration)

    @property
    def account_info(self):
        """Return account information."""
        return lxml.etree.tostring(self._request("account_info")[1][0])

    def _request(self, command, **payload):
        """Send an API request."""
        payload.update(command=command, key=self.token)
        response = requests.get(self.endpoint, params=payload)
        message = lxml.etree.fromstring(response.text)
        try:
            if message.cssselect("ResponseCode")[0].text == "-1":
                print(response.text)
                raise TokenError()
        except IndexError:
            pass
        return message


class NameCom(Registrar):
    """
    [Name.com][0] client.

    [0]: https://name.com

    """

    endpoint = "https://api.name.com"

    def __init__(self, username=None, token=None):
        """Store the username."""
        self.username = username
        super().__init__(token)

    def list_domains(self):
        """List currently registered domains."""
        return [
            (domain["domainName"], domain["expireDate"])
            for domain in self._request("get", "domains")["domains"]
        ]

    def create_record(self, domain, record, subdomain=""):
        """
        Set DNS record for given domain.

        https://www.name.com/api-docs/DNS#CreateRecord

        """
        return self._request(
            "post",
            f"domains/{domain}/records",
            host=subdomain,
            type="A",
            answer=record,
            ttl="300",
        )

    def _request(self, method, command, **payload):
        """Send an API request."""
        post_body = {}
        if payload:
            post_body = {"json": payload}
        response = getattr(requests, method)(
            f"{self.endpoint}/v4/{command}",
            auth=(self.username, self.token),
            headers={"Content-Type": "application/json"},
            **post_body,
        )
        return response.json()



"""A simple Let's Encrypt interface (via `acme-tiny`)."""

import pathlib
import subprocess

import acme_tiny


def generate_cert(domain, certs_dir="certs"):
    """Generate a TLS certificate signed by Let's Encrypt for given domain."""
    certs_dir = pathlib.Path(certs_dir)
    challenge_dir = certs_dir / "challenges"
    challenge_dir.mkdir(exist_ok=True, parents=True)

    account_key = certs_dir / "account.key"
    if not account_key.exists():
        with account_key.open("w") as fp:
            subprocess.call(["openssl", "genrsa", "4096"], stdout=fp)

    domain_dir = certs_dir / domain
    domain_dir.mkdir(exist_ok=True)
    private_key = domain_dir / "domain.key"
    if not private_key.exists():
        with private_key.open("w") as fp:
            subprocess.call(["openssl", "genrsa", "4096"], stdout=fp)
    csr = domain_dir / "domain.csr"
    if not csr.exists():
        with csr.open("w") as fp:
            subprocess.call(
                [
                    "openssl",
                    "req",
                    "-new",
                    "-sha256",
                    "-key",
                    private_key,
                    "-subj",
                    "/",
                    "-addext",
                    f"subjectAltName = DNS:{domain}, DNS:www.{domain}",
                ],
                stdout=fp,
            )
    with (domain_dir / "domain.crt").open("w") as fp:
        fp.write(acme_tiny.get_crt(account_key, csr, challenge_dir))



import json
import pathlib
import subprocess

import dns.exception
import dns.resolver
import easyuri
import web
import whois

from .. import providers

app = web.application(__name__)
config_file = pathlib.Path(".webapp").expanduser()

domain_registrars = {"101domain grs ltd": "101domain.com"}
domain_hosts = {
    "101domain.com": (
        "101domain",
        "101domain.com",
        "help.101domain.com/domain-management/name-servers-dns/modifying-name-servers-and-records/managing-name-server-records",
    ),
    "domaincontrol.com": (
        "GoDaddy",
        "godaddy.com",
        "www.godaddy.com/help/add-an-a-record-19238",
    ),
    "dreamhost.com": (
        "DreamHost",
        "dreamhost.com",
        "help.dreamhost.com/hc/en-us/articles/360035516812-Adding-custom-DNS-records",
    ),
    "dyna-ns.net": (
        "Dynadot",
        "dynadot.com",
        "www.dynadot.com/community/help/question/create-A-record",
    ),
    "name.com": (
        "Name.com",
        "name.com",
        "www.name.com/support/articles/205188538-pointing-your-domain-to-hosting-with-a-records",
    ),
}


def get_config():
    """Get configuration."""
    try:
        with config_file.open() as fp:
            config = json.load(fp)
    except FileNotFoundError:
        config = {}
    return config


def update_config(**items):
    """Update configuration."""
    config = get_config()
    config.update(**items)
    with config_file.open("w") as fp:
        json.dump(config, fp, indent=2, sort_keys=True)
        fp.write("\n")
    return get_config()


def get_onions():
    onions = []
    for hidden_service in pathlib.Path("/var/lib/tor").glob("main*"):
        with (hidden_service / "hostname").open() as fp:
            onions.append(fp.read().strip())
    return onions


@app.wrap
def contextualize(handler, mainapp):
    web.tx.host.local_ip = (
        subprocess.run(["hostname", "-I"], capture_output=True)
        .stdout.split()[0]
        .decode()
    )
    yield


@app.control("")
class Index:
    def get(self):
        config = get_config()
        return app.view.index(
            config.get("application"),
            get_onions(),
            config.get("domains", []),
        )


@app.control("application")
class Application:
    def get(self):
        return app.view.application(get_config().get("application"))


@app.control("onions")
class Onions:
    def get(self):
        return app.view.onions(get_onions())


@app.control("domains")
class Domains:
    def get(self):
        return app.view.domains(get_config().get("domains", []))


class A:
    def get(self):
        form = web.form(name=None, domain=None)
        try:
            with open("/var/lib/tor/main/hostname") as fp:
                onion = fp.read().strip()
        except FileNotFoundError:
            onion = "s9d8df3jhif98.onion"
        local_ip = (
            subprocess.run(["hostname", "-I"], capture_output=True)
            .stdout.split()[0]
            .decode()
        )
        domain_ns = None
        domain_ip = None
        www_domain_ip = None
        domain_whois = None
        domain_host = None
        if form.domain:
            resolver = dns.resolver.Resolver()
            # resolver.nameservers = ["8.8.8.8"]
            try:
                domain_ip = list(resolver.resolve(form.domain).rrset)[0].address
                www_domain_ip = list(resolver.resolve(f"www.{form.domain}").rrset)[
                    0
                ].address
            except dns.exception.DNSException:
                pass
            domain_whois = whois.whois(form.domain)
            try:
                if domain_whois["name_servers"]:
                    _nameservers = sorted(set(domain_whois["name_servers"]))
                    domain_host = domain_hosts[
                        easyuri.parse(_nameservers[0]).suffixed_domain
                    ]
                else:
                    domain_host = domain_hosts[
                        domain_registrars[domain_whois["registrar"].lower()]
                    ]
            except KeyError:
                pass
            if local_ip == domain_ip and form.action == "bootstrap":
                return "bootstrap"
        return app.view.index(
            onion,
            form.name,
            local_ip,
            form.domain,
            domain_whois,
            domain_host,
            domain_ns,
            domain_ip,
            www_domain_ip,
        )



"""
HTTP headers

"""

from . import entity, general, request, response

__all__ = ["get_header", "Headers"]


def get_header(header, value=None):
    """ """
    header = header.lower().replace("_", "").replace("-", "")
    for header_type in (entity, general, request, response):
        for search_name in header_type.__all__:
            if search_name.lower() == header:
                return getattr(header_type, search_name)(value)
    return value


def _title_case(header):
    """

    >>> _title_case("accept_Encoding")
    'Accept-Encoding'

    """
    # TODO re.split on -, _, whitespace, or camelcase
    return "-".join(part.capitalize() for part in header.split("_"))


class Headers(dict):

    """
    a dot-mapping optimized to hold HTTP header objects

        >>> h = Headers(**{"content_type": "text/plain", "CONTENT-LENGTH": "555"})
        >>> h["content-type"]
        text/plain
        >>> h["content-length"]
        555
        >>> h.content_type
        text/plain
        >>> type(h.content_type)
        <class 'web.headers.entity.ContentType'>

    """

    def __init__(self, **headers):
        for header, value in headers.items():
            self[header] = value

    @classmethod
    def from_lines(cls, lines):
        """
        return a `Headers` object based upon iterable of `header` strings

            >>> header_lines = '''
            ... Date: Mon, 27 Jul 2009 12:28:53 GMT
            ... Server: Apache
            ... Last-Modified: Wed, 22 Jul 2009 19:15:56 GMT
            ... ETag: "34aa387-d-1568eb00"
            ... Accept-Ranges: bytes
            ... Content-Length: 14
            ... Vary: Accept-Encoding
            ... Content-Type: text/plain
            ...
            ... '''
            >>> headers = Headers.from_lines(header_lines)

        """
        headers = cls()
        for line in lines.splitlines():
            name, _, value = (part.strip() for part in line.partition(":"))
            headers[name] = value
        return headers

    # @property
    # def request(self):
    #     headers = {}
    #     for k, v in self.__dict__.items():
    #         if issubclass(v.__class__, (Request, Entity, General)):
    #             headers[_title_case(k)] = str(v)
    #     return headers

    @property
    def response(self):
        # FIXME Set-Cookie broke this function
        headers = {}
        for k, v in self.__dict__.items():
            # if v.__class__.__name__ in ["SetCookie"]:
            #     headers[_title_case(k)] = v.header
            if issubclass(
                v.__class__, (response.Response, entity.Entity, general.General)
            ):
                headers[_title_case(k)] = str(v)
        return headers

    @property
    def wsgi(self):
        headers = []
        for name, value in self.__dict__.items():
            name = _title_case(name)
            # XXX if name in ["Set-Cookie"]:
            # XXX     headers.extend((name, val) for val in value.header)
            # XXX else:
            if isinstance(value, list):
                for item in value:
                    headers.append((name, str(item)))
            else:
                headers.append((name, str(value)))
        return headers

    def get(self, header, default=None):
        return getattr(self, header, default)

    def __getitem__(self, header):
        header = header.lower().replace("_", "-")
        return super(Headers, self).__getitem__(header)

    def __setitem__(self, header, value):
        header = header.lower().replace("_", "-")
        obj = get_header(header, value)
        self.__dict__[header] = obj
        super(Headers, self).__setitem__(header, obj)

    def __getattr__(self, header):
        try:
            return self[header]
        except KeyError:
            raise AttributeError(header)

    def __setattr__(self, header, value):
        self[header] = value

    def __contains__(self, header):
        header = header.lower().replace("_", "-")
        return header in self.__dict__

    # def __repr__(self):
    #     return repr({_title_case(k): v for k, v in sorted(self.items())})



"""
A color palette that adapts to the time of day.

Solarized is a sixteen color palette (eight monotones and eight accents)
designed with both precise CIELAB lightness relationships and a refined
set of hues based on fixed color wheel relationships. It is intended for
use with terminal and GUI applications.

The [Solarized][1] color palette is designed by @[Ethan Schoonover][2].

[1]: http://ethanschoonover.com/solarized
[2]: http://ethanschoonover.com

Styles based upon the original implementation `pygments-style-solarized`
by @[Shoji Kumagai](mailto:take.this.2.your.grave@gmail.com) and released
under the MIT license.

"""

# TODO import uuid
import re

import lxml.html
import pygments
import pygments.formatters
import pygments.lexers
import pygments.style
from lxml.html import builder
from pygments.token import Literal  # Token
from pygments.token import (Comment, Error, Generic, Keyword, Name, Number,
                            Operator, Other, Punctuation, String, Text,
                            Whitespace)

__all__ = ["highlight", "Solar", "Lunar"]


bases = _b = {
    "03": "002b36",
    "0": "839496",
    "02": "073642",
    "1": "93a1a1",
    "01": "586e75",
    "2": "eee8d5",
    "00": "657b83",
    "3": "fdf6e3",
}
dark_bases = (_b["03"], _b["02"], _b["01"], _b["0"], _b["1"])
light_bases = (_b["3"], _b["2"], _b["0"], _b["00"], _b["01"])
colors = _c = {
    "yellow": "b58900",
    "orange": "cb4b16",
    "red": "dc322f",
    "magenta": "d33682",
    "violet": "6c71c4",
    "blue": "268bd2",
    "cyan": "2aa198",
    "green": "859900",
}
palette = tuple(list(bases.values()) + list(colors.values()))
accented = [
    (_b["02"], _b["01"], _b["0"], _b["1"], _c["red"]),
    (_b["03"], _b["02"], _b["01"], _b["00"], _c["yellow"]),
    (_b["00"], _b["1"], _b["2"], _b["3"], _c["blue"]),
    (_b["00"], _b["0"], _b["1"], _b["2"], _c["orange"]),
    (_b["02"], _b["1"], _b["2"], _c["magenta"], _c["violet"]),
]


def highlight(
    code,
    filename,
    lines=True,
    starting_line=1,
    focus=None,
    coverage=None,
    failures=None,
    search=None,
    diff=None,
):
    """
    Highlight given code.

    Focus any lines containing `search`.

    """
    if filename.endswith(".py"):
        lexer = pygments.lexers.Python3Lexer()
    else:
        try:
            lexer = pygments.lexers.get_lexer_for_filename(filename)
        except pygments.util.ClassNotFound:
            lexer = pygments.lexers.get_lexer_for_filename(".txt")
            # XXX return code
    lexer.add_filter("codetagify")
    # TODO random_id = uuid.uuid4()
    is_file = filename.rpartition(".")[0]  # an actual filename, eg "date_util.py"
    if not is_file:
        filename = f"XaXaX{filename}"
    # if diff:
    #     starting_line = diff[0]
    starting_line = int(starting_line)
    if diff:
        formatter_args = {}
    else:
        formatter_args = {"linenos": "table", "linenostart": starting_line}
    formatter = pygments.formatters.HtmlFormatter(linespans=filename, **formatter_args)
    html = pygments.highlight(code, lexer, formatter)

    doc = lxml.html.fromstring(html)
    if search:
        focus = []
        for line in doc.cssselect("td.code > div > pre > span"):
            if re.search(search, line.text_content()):
                focus.append(line.attrib["id"].rpartition("-")[2])

    if focus:
        if isinstance(focus, int):
            focus = [focus]
        for focus_line in focus:
            # XXX given_id = f"{random_id}-L{focus_line}".format(random_id, focus_line)
            given_id = str(focus_line)
            html = html.replace(given_id, f'{given_id}" class="focus')
    if coverage:
        for line, covered in coverage:
            if int(covered):
                class_name = "covered"
            else:
                class_name = "uncovered"
            current_id = f'{filename}-{line}"'
            html = html.replace(current_id, f'{current_id} class="{class_name}"')
    if failures:
        for line in failures:
            current_id = f"{filename}-L{line}"
            html = html.replace(current_id, f'{current_id}" class="failed')

    if diff:
        ...
    else:
        doc = lxml.html.fromstring(html)
        linenos = doc.cssselect(".linenodiv pre")[0]
        ending_line = linenos.text_content().rpartition("\n")[2].partition('"')[0]
        for lineno in linenos.iterchildren():
            lineno.drop_tree()
        linenos.text = ""
        for line in range(starting_line, int(ending_line) + 1):
            line_link = f"#L{line}"
            if is_file:
                line_link = f"#{filename}-L{line}"
            linenos.append(builder.A(str(line), "\n", href=line_link))
        for span in doc.cssselect(".highlight pre > span"):
            if span.text:
                span.text = span.text.replace(" ", "\u00a0")
        html = lxml.html.tostring(doc).decode("utf-8")
        if not is_file:
            html = html.replace(f"{filename}-", "L")
    return html


class Solar(pygments.style.Style):
    """A Pygments style based upon Solarized's light palette."""

    background_color = "#fdf6e3"
    default_style = ""

    styles = {
        Text: "#657b83",  # base00
        Whitespace: "#fdf6e3",  # base3   w
        Error: "#dc322f",  # red     err
        Other: "#657b83",  # base00  x
        Comment: "italic #93a1a1",  # base1   c
        Comment.Multiline: "italic #93a1a1",  # base1   cm
        Comment.Preproc: "italic #93a1a1",  # base1   cp
        Comment.Single: "italic #93a1a1",  # base1   c1
        Comment.Special: "italic #93a1a1",  # base1   cs
        Keyword: "#859900",  # green   k
        Keyword.Constant: "#859900",  # green   kc
        Keyword.Declaration: "#859900",  # green   kd
        Keyword.Namespace: "#cb4b16",  # orange  kn
        Keyword.Pseudo: "#cb4b16",  # orange  kp
        Keyword.Reserved: "#859900",  # green   kr
        Keyword.Type: "#859900",  # green   kt
        Operator: "#657b83",  # base00  o
        Operator.Word: "#859900",  # green   ow
        Name: "#586e75",  # base01  n
        Name.Attribute: "#657b83",  # base00  na
        Name.Builtin: "#268bd2",  # blue    nb
        Name.Builtin.Pseudo: "bold #268bd2",  # blue    bp
        Name.Class: "#268bd2",  # blue    nc
        Name.Constant: "#b58900",  # yellow  no
        Name.Decorator: "#cb4b16",  # orange  nd
        Name.Entity: "#cb4b16",  # orange  ni
        Name.Exception: "#cb4b16",  # orange  ne
        Name.Function: "#268bd2",  # blue    nf
        Name.Property: "#268bd2",  # blue    py
        Name.Label: "#657b83",  # base00  nc
        Name.Namespace: "#b58900",  # yellow  nn
        Name.Other: "#657b83",  # base00  nx
        Name.Tag: "#859900",  # green   nt
        Name.Variable: "#cd4b16",  # orange  nv
        Name.Variable.Class: "#268bd2",  # blue    vc
        Name.Variable.Global: "#268bd2",  # blue    vg
        Name.Variable.Instance: "#268bd2",  # blue    vi
        Number: "#2aa198",  # cyan    m
        Number.Float: "#2aa198",  # cyan    mf
        Number.Hex: "#2aa198",  # cyan    mh
        Number.Integer: "#2aa198",  # cyan    mi
        Number.Integer.Long: "#2aa198",  # cyan    il
        Number.Oct: "#2aa198",  # cyan    mo
        Literal: "#657b83",  # base00  l
        Literal.Date: "#657b83",  # base00  ld
        Punctuation: "#657b83",  # base00  p
        String: "#2aa198",  # cyan    s
        String.Backtick: "#2aa198",  # cyan    sb
        String.Char: "#2aa198",  # cyan    sc
        String.Doc: "#2aa198",  # cyan    sd
        String.Double: "#2aa198",  # cyan    s2
        String.Escape: "#cb4b16",  # orange  se
        String.Heredoc: "#2aa198",  # cyan    sh
        String.Interpol: "#cb4b16",  # orange  si
        String.Other: "#2aa198",  # cyan    sx
        String.Regex: "#2aa198",  # cyan    sr
        String.Single: "#2aa198",  # cyan    s1
        String.Symbol: "#2aa198",  # cyan    ss
        Generic: "#657b83",  # base00  g
        Generic.Deleted: "#657b83",  # base00  gd
        Generic.Emph: "#657b83",  # base00  ge
        Generic.Error: "#657b83",  # base00  gr
        Generic.Heading: "#657b83",  # base00  gh
        Generic.Inserted: "#657b83",  # base00  gi
        Generic.Output: "#657b83",  # base00  go
        Generic.Prompt: "#657b83",  # base00  gp
        Generic.Strong: "#657b83",  # base00  gs
        Generic.Subheading: "#657b83",  # base00  gu
        Generic.Traceback: "#657b83",  # base00  gt
    }


class Lunar(pygments.style.Style):
    """A Pygments style based upon Solarized's dark palette."""

    background_color = "#002b36"
    default_style = ""

    styles = {
        Text: "#839496",  # base0
        Whitespace: "#002b36",  # base03 w
        Error: "#dc322f",  # red    err
        Other: "#839496",  # base0  x
        Comment: "italic #586e75",  # base01 c
        Comment.Multiline: "italic #586e75",  # base01 cm
        Comment.Preproc: "italic #586e75",  # base01 cp
        Comment.Single: "italic #586e75",  # base01 c1
        Comment.Special: "italic #586e75",  # base01 cs
        Keyword: "#859900",  # green  k
        Keyword.Constant: "#859900",  # green  kc
        Keyword.Declaration: "#859900",  # green  kd
        Keyword.Namespace: "#cb4b16",  # orange kn
        Keyword.Pseudo: "#cb4b16",  # orange kp
        Keyword.Reserved: "#859900",  # green  kr
        Keyword.Type: "#859900",  # green  kt
        Operator: "#839496",  # base0  o
        Operator.Word: "#859900",  # green  ow
        Name: "#93a1a1",  # base1  n
        Name.Attribute: "#839496",  # base0  na
        Name.Builtin: "#268bd2",  # blue   nb
        Name.Builtin.Pseudo: "bold #268bd2",  # blue   bp
        Name.Class: "#268bd2",  # blue   nc
        Name.Constant: "#b58900",  # yellow no
        Name.Decorator: "#cb4b16",  # orange nd
        Name.Entity: "#cb4b16",  # orange ni
        Name.Exception: "#cb4b16",  # orange ne
        Name.Function: "#268bd2",  # blue   nf
        Name.Property: "#268bd2",  # blue   py
        Name.Label: "#839496",  # base0  nc
        Name.Namespace: "#b58900",  # yellow nn
        Name.Other: "#839496",  # base0  nx
        Name.Tag: "#859900",  # green  nt
        Name.Variable: "#cd4b16",  # orange nv
        Name.Variable.Class: "#268bd2",  # blue   vc
        Name.Variable.Global: "#268bd2",  # blue   vg
        Name.Variable.Instance: "#268bd2",  # blue   vi
        Number: "#2aa198",  # cyan   m
        Number.Float: "#2aa198",  # cyan   mf
        Number.Hex: "#2aa198",  # cyan   mh
        Number.Integer: "#2aa198",  # cyan   mi
        Number.Integer.Long: "#2aa198",  # cyan   il
        Number.Oct: "#2aa198",  # cyan   mo
        Literal: "#839496",  # base0  l
        Literal.Date: "#839496",  # base0  ld
        Punctuation: "#839496",  # base0  p
        String: "#2aa198",  # cyan   s
        String.Backtick: "#2aa198",  # cyan   sb
        String.Char: "#2aa198",  # cyan   sc
        String.Doc: "#2aa198",  # cyan   sd
        String.Double: "#2aa198",  # cyan   s2
        String.Escape: "#cb4b16",  # orange se
        String.Heredoc: "#2aa198",  # cyan   sh
        String.Interpol: "#cb4b16",  # orange si
        String.Other: "#2aa198",  # cyan   sx
        String.Regex: "#2aa198",  # cyan   sr
        String.Single: "#2aa198",  # cyan   s1
        String.Symbol: "#2aa198",  # cyan   ss
        Generic: "#839496",  # base0  g
        Generic.Deleted: "#839496",  # base0  gd
        Generic.Emph: "#839496",  # base0  ge
        Generic.Error: "#839496",  # base0  gr
        Generic.Heading: "#839496",  # base0  gh
        Generic.Inserted: "#839496",  # base0  gi
        Generic.Output: "#839496",  # base0  go
        Generic.Prompt: "#839496",  # base0  gp
        Generic.Strong: "#839496",  # base0  gs
        Generic.Subheading: "#839496",  # base0  gu
        Generic.Traceback: "#839496",  # base0  gt
    }



import shutil
from pathlib import Path

from . import Machine

test_dir = Path("/tmp/webint-web.host-test")


def setup_module(module):
    try:
        shutil.rmtree(test_dir)
    except FileNotFoundError:
        pass


def test_machine():
    machine = Machine()
    assert machine.run(f"mkdir {test_dir}").returncode == 0
    assert machine.run(f"ls {test_dir} -1").returncode == 0


def test_cd():
    machine = Machine()
    with machine.cd(test_dir) as root_dir:
        root_dir.run("mkdir foobar").returncode == 0
        with root_dir.cd("foobar") as foobar_dir:
            foobar_dir.run("mkdir batbaz").returncode == 0
            with foobar_dir.cd("batbaz") as batbaz_dir:
                batbaz_dir.run("mkdir a b c").returncode == 0
                assert batbaz_dir.run("ls -1").lines[:3] == ["a", "b", "c"]
    assert (test_dir / "foobar/batbaz/a").exists()



"""

"""

from .util import Status

__all__ = ["ServerError", "InternalServerError", "NotImplemented", "ServiceUnavailable"]


class ServerError(Status):

    """5xx -- the server failed to fulfill an apparently valid request"""


class InternalServerError(ServerError):

    """
    500

    """


class NotImplemented(ServerError):

    """
    501

    """


class ServiceUnavailable(ServerError):

    """
    503

    """



"""
Tools for a metamodern web environment.

## User agent tools

Simple interface, simple automate.

## Web application framework

Simple interface, simple deploy.

"""

import sys

if len(sys.argv) > 1 and sys.argv[1] != "run":
    from gevent import monkey

    monkey.patch_all()

import mf
from dns import resolver as dns
from webagt import ConnectionError, get

from . import host
from .framework import *  # noqa
from .markdown import render as mkdn
from .response import Status  # noqa
from .response import (OK, Accepted, BadRequest, Conflict, Created, Forbidden,
                       Found, Gone, MethodNotAllowed, MovedPermanently,
                       MultiStatus, NoContent, NotFound, PermanentRedirect,
                       SeeOther, Unauthorized)
from .templating import template, templates

__all__ = [
    "mf",
    "dns",
    "mkdn",
    "get",
    "ConnectionError",
    "template",
    "templates",
    "OK",
    "Accepted",
    "BadRequest",
    "Conflict",
    "Created",
    "Forbidden",
    "Found",
    "Gone",
    "MethodNotAllowed",
    "MultiStatus",
    "NoContent",
    "NotFound",
    "PermanentRedirect",
    "SeeOther",
    "Unauthorized",
]


abba = application("foobar")


@abba.control("")
class Front:
    def get(self):
        return "fnord"



"""Command line tools for the web."""

import inspect
import json
import logging
import pathlib
import time
import webbrowser

import gevent
import txt
import webagt
from rich.pretty import pprint

import web
import web.host
from web.host import console, providers

__all__ = ["main"]


main = txt.application("web", web.__doc__)
config_file = pathlib.Path("~/.webint").expanduser()


def get_config():
    """Get configuration."""
    try:
        with config_file.open() as fp:
            config = json.load(fp)
    except FileNotFoundError:
        config = {}
    return config


def update_config(**items):
    """Update configuration."""
    config = get_config()
    config.update(**items)
    with config_file.open("w") as fp:
        json.dump(config, fp, indent=2, sort_keys=True)
        fp.write("\n")
    return get_config()


@main.register()
class Apps:
    """Show installed web apps."""

    def run(self, stdin, log):
        for pkg, apps in web.get_apps().items():
            for name, _, ns, obj in apps:
                print(f"{name} {ns}:{obj[0]}")
        return 0


@main.register()
class Fnord:
    def run(self, stdin, log):
        # import asyncio
        # asyncio.run(web.serve("web:abba", port=9999))
        web.StandaloneServer(web.abba, 9999).run()


@main.register()
class Run:
    """Run a web app locally."""

    def setup(self, add_arg):
        add_arg("app", help="name of web application")
        add_arg("--port", help="port to serve on")
        add_arg("--socket", help="file socket to serve on")
        add_arg("--watch", default=".", help="directory to watch for changes")

    def run(self, stdin, log):
        import asyncio

        if self.port:
            asyncio.run(web.serve(self.app, port=self.port, watch_dir=self.watch))
        elif self.socket:
            asyncio.run(web.serve(self.app, socket=self.socket, watch_dir=self.watch))
        else:
            print("must provide a port or a socket")
            return 1
        return 0

        # for pkg, apps in web.get_apps().items():
        #     for name, _, ns, obj in apps:
        #         if self.app == name:
        #             web.serve(ns, obj)
        #             return 0
        # return 1


def get_providers(provider_type):
    return [
        name.lower()
        for name, obj in inspect.getmembers(providers, inspect.isclass)
        if issubclass(obj, provider_type) and obj is not provider_type
    ]


@main.register()
class Config:
    """Config your environments."""

    def setup(self, add_arg):
        add_arg("token", help="API access token")
        add_arg(
            "--host",
            choices=get_providers(providers.Host),
            help="machine host",
        )
        add_arg(
            "--registrar",
            choices=get_providers(providers.Registrar),
            help="domain registrars",
        )

    def run(self, stdin, log):
        if self.host:
            update_config(host=self.host, host_token=self.token)
        elif self.registrar:
            update_config(registrar=self.registrar, registrar_token=self.token)
        return 0


@main.register()
class Bootstrap:
    """Bootstrap a website."""

    def setup(self, add_arg):
        add_arg("website", help="name of website")
        add_arg("package", help="name of PyPI package to install")
        add_arg("app", help="name of web app to run")

    def run(self, stdin, log):
        logging.basicConfig(
            level=logging.DEBUG,
            filename="debug.log",
            filemode="w",
            force=True,
            format="%(levelname)s:%(asctime)s:%(name)s:%(message)s",
        )
        config = get_config()

        if config["host"] == "digitalocean":
            host = providers.DigitalOcean(config["host_token"])
        elif config["host"] == "linode":
            host = providers.Linode(config["host_token"])
        else:
            console.print(f"Host {config['host']} not available.")
            return 1

        # if config["registrar"] == "dynadot":
        #     registrar = providers.Dynadot(config["registrar_token"])
        # elif config["registrar"] == "linode":
        #     registrar = providers.Linode(config["registrar_token"])
        # else:
        #     console.print(f"Registrar {config['registrar']} not available.")
        #     return 1

        secret = web.nbrandom(4)
        machines = config.get("machines", {})
        versions = web.host.Machine.versions
        start_time = time.time()
        total_time = 14
        with console.status(
            f"[bold green]around {total_time} minutes remaining"
        ) as status:

            def update_status():
                while True:
                    remaining_time = total_time - round((time.time() - start_time) / 60)
                    if remaining_time < 1:
                        status.update("Finishing up..")
                        return
                    status.update(
                        f"[bold green]around {remaining_time} minutes remaining"
                    )
                    gevent.sleep(1)

            gevent.spawn(update_status)

            console.print(f"Spawning virtual machine at {host.__class__.__name__}")
            machine = web.host.spawn_machine(self.website, host)
            console.print("Updating system")
            machine._apt("update")
            console.print("Installing firewall")
            machine._apt("install -yq ufw")
            machine.open_ports(22)
            console.print("Setting up system")
            machine.setup_machine()
            console.print("Installing python")
            machine.setup_python()
            console.print("Installing tor")
            machine.setup_tor()
            # XXX machine = web.host.Machine("128.199.1.111", "root", "gaea_key")
            # XXX machine.onion = (
            # XXX     "hxpcvad6txrivhondry6htgmlc2idrxz37ofyobe7ynxpdxd7xdn5sid.onion"
            # XXX )
            console.print("Installing nginx")
            machine.setup_nginx()
            console.print(f"Installing application: {self.app} of {self.package}")
            machine.setup_app(self.package, self.app, config["host_token"], secret)
            time.sleep(2)
        console.rule(
            f"[green]Bootstrapped {self.package} ({self.app}) at {machine.address}"
        )
        webbrowser.open(f"https://{machine.address}?secret={secret}")
        machines[self.website] = machine.address
        update_config(machines=machines)
        return 0


@main.register()
class Host:
    """Manage your host."""

    def run(self, stdin, log):
        config = get_config()
        if config["host"] == "digitalocean":
            host = providers.DigitalOcean(config["host_token"])
        else:
            console.print(f"Host {config['host']} not available.")
            return 1
        for machine in host.machines:
            console.rule(f"[bold red]{machine['name']}")
            pprint(machine)
        return 0


@main.register()
class Registrar:
    """Manage your registrar."""

    def run(self, stdin, log):
        config = get_config()
        if config["registrar"] == "dynadot":
            registrar = providers.Dynadot(config["registrar_token"])
        else:
            console.print(f"Registrar {config['registrar']} not available.")
            return 1
        for domain in registrar.domains:
            print(domain)
        return 0


if __name__ == "__main__":
    main()

# nuitka-project: --include-package=gevent.signal
# nuitka-project: --include-package=gunicorn.glogging
# nuitka-project: --include-package=gunicorn.workers.sync
# nuitka-project: --include-package=web.framework.templates
# nuitka-project: --include-package=web.host.templates
## nuitka-project: --include-package-data=mf2py
## nuitka-project: --include-package-data=selenium
